#include "memory_preallocated.h"

#include <stdlib.h>

/* Initialise the memory pool with a mandatory 'limit' in pages. */

MemoryPreallocated::MemoryPreallocated(unsigned int limit, offset_t region_size)
: Memory(region_size), _limit(limit)
{
    /* Allocate all memory at once, rounding each unit to the nearest page
       size. */

    offset_t rounded_size = round_multiple(region_size, PAGE_SIZE);
    offset_t total_size = limit * rounded_size;
    offset_t current;

    if (posix_memalign((void **) &current, total_size, total_size))
        return;

    /* Split the memory up into separate regions. */

    while (limit)
    {
        _free.push_back(new Region(current, current + rounded_size));
        current += rounded_size;
        limit--;
    }
}

/* Free all allocated regions. */

MemoryPreallocated::~MemoryPreallocated()
{
    for (std::list<Region *>::iterator it = _free.begin(); it != _free.end(); it++)
        delete *it;

    for (std::set<Region *>::iterator it = _used.begin(); it != _used.end(); it++)
        delete *it;
}

/* Return whether memory has been allocated. */

bool MemoryPreallocated::allocated()
{
    std::lock_guard<std::mutex> guard(_lock);

    return !(_free.empty() && _used.empty());
}

/* Allocate a new region of the configured size rounded to the nearest page. */

Region *MemoryPreallocated::region()
{
    std::lock_guard<std::mutex> guard(_lock);

    std::list<Region *>::iterator it = _free.begin();

    if (it == _free.end())
        return NULL;

    Region *r = *it;

    _free.erase(it);
    _used.insert(r);

    return r;
}

/* Release the allocated 'region'. */

void MemoryPreallocated::release(Region *region)
{
    std::lock_guard<std::mutex> guard(_lock);

    std::set<Region *>::iterator it = _used.find(region);

    if (it != _used.end())
    {
        _used.erase(it);
        _free.push_back(region);
    }
}

// vim: tabstop=4 expandtab shiftwidth=4
